# -*- coding: utf-8 -*-
##############################################################################
#
#    OmniaSolutions, Your own solutions
#    Copyright (C) 2010 OmniaSolutions (<http://omniasolutions.eu>). All Rights Reserved
#    $Id$
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import os
import time
import types
from datetime import datetime
import logging

from openerp import api
from openerp.osv import osv, fields
from openerp.tools.translate import _

USED_STATES=[('draft','Draft'),('confirmed','Confirmed'),('released','Released'),('undermodify','UnderModify'),('obsoleted','Obsoleted')]
USEDIC_STATES=dict(USED_STATES)
#STATEFORRELEASE=['confirmed']
#STATESRELEASABLE=['confirmed','transmitted','released','undermodify','obsoleted']

class plm_component(osv.osv):
    _name = 'product.product'
    _inherit = 'product.product'
    _columns = {
                'create_date': fields.datetime('Date Created', readonly=True),
                'write_date': fields.datetime('Date Modified', readonly=True),
    }

#   Internal methods

    @api.model
    def createCuttedPartsBOM(self, bom_structure):
        pass
        
    def _getbyrevision(self, cr, uid, name, revision):
        result=None
        results=self.search(cr,uid,[('engineering_code','=',name),('engineering_revision','=',revision)])
        for result in results:
            break
        return result

    def getRecursiveParts(self, cr, uid, prodId, excludeStates=[], context=None):
        prodEnv = self.pool.get('product.product')
        prodBrws = prodEnv.browse(cr, uid, prodId)
        return self.getRecursivePartsBom(cr, uid, prodBrws, excludeStates, level=0, outLevelDict={}, alreadyEvaluated=[], context={'skip': True})

    def getRecursiveInfos(self, prodBrws, excludeStates=[], alreadyEvaluated=[]):
        if prodBrws.state not in excludeStates:
            outDocs = []
            componentKey = 'c_%s' % (prodBrws.id)
            if componentKey in alreadyEvaluated:
                return
            for documentBrws in prodBrws.linkeddocuments:
                documentKey = 'd_%s' % (documentBrws.id)
                filename = documentBrws.datas_fname
                if documentBrws.state not in excludeStates and not filename.upper().endswith('.PDF'):
                    if documentKey not in alreadyEvaluated:
                        outDocs.append({
                            'id': documentBrws.id,
                            'name': documentBrws.name,
                            'revisionid': documentBrws.revisionid,
                            'description': documentBrws.description,
                            'state': documentBrws.state,
                            'datas_fname': filename,
                            })
                        alreadyEvaluated.append(documentKey)
            alreadyEvaluated.append(componentKey)
            return {
                'id': prodBrws.id,
                'name': prodBrws.name,
                'engineering_code': prodBrws.engineering_code,
                'engineering_revision': prodBrws.engineering_revision,
                'description': prodBrws.description,
                'state': prodBrws.state,
                'documents': outDocs,
                }
        
    def getRecursivePartsBom(self, cr, uid, prodBrws, excludeStates=[], level=0, outLevelDict={}, alreadyEvaluated=[], context=None):
        skipFlag = context.get('skip', False)
        if unicode(level) not in outLevelDict.keys():
            outLevelDict[unicode(level)] = []
        if not skipFlag:
            partInfos = self.getRecursiveInfos(prodBrws, excludeStates, alreadyEvaluated)
            if partInfos:
                outLevelDict[unicode(level)].append(partInfos)
        context['skip'] = False
        for bomBrws in prodBrws.product_tmpl_id.bom_ids:
            for bomlineBrws in bomBrws.bom_line_ids:
                self.getRecursivePartsBom(cr, uid, bomlineBrws.product_id, excludeStates, level=level+1, outLevelDict=outLevelDict, alreadyEvaluated=alreadyEvaluated, context=context)
        return outLevelDict
            
        
    def _getChildrenBom(self, cr, uid, component, level=0, currlevel=0, context=None):
        """
            Return a flat list of each child, listed once, in a Bom ( level = 0 one level only, level = 1 all levels)
        """
        result=[]
        bufferdata=[]
        if level==0 and currlevel>1:
            return bufferdata
        for bomid in component.product_tmpl_id.bom_ids:
            for bomline in bomid.bom_line_ids:
                children=self._getChildrenBom(cr, uid, bomline.product_id, level, currlevel+1, context=context)
                bufferdata.extend(children)
                bufferdata.append(bomline.product_id.id)
        result.extend(bufferdata)
        return list(set(result))

    def RegMessage(self, cr, uid, request, default=None, context=None):
        """
            Registers a message for requested component
        """
        oid, message = request
        self.wf_message_post(cr, uid, [oid], body=_(message))
        return False

    def getLastTime(self, cr, uid, oid, default=None, context=None):
        return self.getUpdTime(self.browse(cr, uid, oid, context=context))

    def getUpdTime(self, obj):
        if(obj.write_date != False):
            return datetime.strptime(obj.write_date, '%Y-%m-%d %H:%M:%S')
        else:
            return datetime.strptime(obj.create_date, '%Y-%m-%d %H:%M:%S')

##  Customized Automations
    def on_change_name(self, cr, uid, oid, name=False, engineering_code=False):
        outDict = {'value': {}}
        if name:
            results = self.search(cr, uid, [('name', '=', name)])
            if len(results) > 0:
                raise osv.except_osv(_('Update Part Warning'), _("\nPart Name %s already exists.\nClose with OK to reuse, with Cancel to discharge." % (name)))
            if not engineering_code:
                engineering_code = name
                outDict['value'].update({'engineering_code': name})
            if self.search(cr, uid, [('engineering_code', '=', engineering_code)]):
                raise osv.except_osv(_('Update Part Warning'), _("\nPart Number %s already exists.\nClose with OK to reuse, with Cancel to discharge." % (engineering_code)))
        return outDict

    def on_change_engineering_code(self, cr, uid, oid, engineering_code=False):
        if engineering_code:
            if self.search(cr, uid, [('engineering_code', '=', engineering_code)]):
                raise osv.except_osv(_('Update Part Warning'), _("\nPart Number %s already exists.\nClose with OK to reuse, with Cancel to discharge." % (engineering_code)))
        return {'value': {'engineering_code': engineering_code}}
##  External methods
    def Clone(self, cr, uid, oid, context=None, defaults={}):
        """
            create a new revision of the component
        """
        exitValues = {}
        newID = self.copy(cr, uid, oid, defaults, context)
        if newID not in (None, False):
            newEnt = self.browse(cr, uid, newID, context=context)
            exitValues['_id'] = newID
            exitValues['name'] = newEnt.name
            exitValues['engineering_code'] = newEnt.engineering_code
            exitValues['engineering_revision'] = newEnt.engineering_revision  # Always equal to 0 because forced by copy overload method
            exitValues['engineering_writable'] = True
            exitValues['state'] = 'draft'
        return exitValues

    def GetUpdated(self,cr,uid,vals,context=None):
        """
            Get Last/Requested revision of given items (by name, revision, update time)
        """
        partData, attribNames, forceCADProperties = vals
        ids = self.GetLatestIds(cr, uid, partData, context, forceCADProperties=forceCADProperties)
        return self.read(cr, uid, list(set(ids)), attribNames)

    def GetLatestIds(self, cr, uid, vals, context=None, forceCADProperties=False):
        """
            Get Last/Requested revision of given items (by name, revision, update time)
        """
        ids = []
        plmDocObj = self.pool.get('plm.document')

        def getCompIds(partName, partRev):
            if partRev is None or partRev is False:
                partIds = self.search(cr, uid, [('engineering_code', '=', partName)], order='engineering_revision', context=context)
                if len(partIds) > 0:
                    partIds.sort()
                    ids.append(partIds[len(partIds) - 1])
            else:
                ids.extend(self.search(cr, uid, [('engineering_code', '=', partName), ('engineering_revision', '=', partRev)], context=context))

        for compName, compRev, docIdToOpen in vals:
            checkOutUser = plmDocObj.get_checkout_user(cr, uid, docIdToOpen, context)
            if checkOutUser:
                isMyDocument = plmDocObj.isCheckedOutByMe(cr, uid, docIdToOpen, context)
                if isMyDocument and forceCADProperties:
                    return []    # Document properties will be not updated
            getCompIds(compName, compRev)
        return list(set(ids))
    
    def NewRevision(self,cr,uid,ids,context=None):
        """
            create a new revision of current component
        """
        newID = None
        newIndex = 0
        logging.info('Start New revision of ids %r' % (ids))
        for tmpObject in self.browse(cr, uid, ids, context=context):
            latestIDs = self.GetLatestIds(cr, uid, [(tmpObject.engineering_code,tmpObject.engineering_revision,False)], context=context)
            logging.info('Latest revision ids %r' % (latestIDs))
            for oldObject in self.browse(cr, uid, latestIDs, context=context):
                newIndex = int(oldObject.engineering_revision) + 1
                defaults = {}
                defaults['engineering_writable'] = False
                defaults['state'] = 'undermodify'
                self.write(cr, uid, [oldObject.id], defaults, context=context)
                self.wf_message_post(cr, uid, [oldObject.id], body=_('Status moved to: %s.' % (USEDIC_STATES[defaults['state']])))
                # store updated infos in "revision" object
                defaults['name'] = oldObject.engineering_code                 # copy function needs an explicit name value
                # ATTENTION!!! field name translatable can do a wrong copy operation
                # Default copy if found a product with the same name return it's id instead of
                # create a new one. So not change in defaults['name'] = oldObject.name
                defaults['engineering_code'] = oldObject.engineering_code
                defaults['engineering_revision'] = newIndex
                defaults['engineering_writable'] = True
                defaults['state'] = 'draft'
                defaults['linkeddocuments'] = []                  # Clean attached documents for new revision object
                logging.info('Going to copy %r with values %r' % (oldObject.id, defaults))
                newID = self.copy(cr, uid, oldObject.id, defaults, context=context)
                self.wf_message_post(cr, uid, [oldObject.id], body=_('Created : New Revision.'))
                break
            break
        return (newID, newIndex) 

    def SaveOrUpdate(self, cr, uid, vals, default=None, context=None):
        """
            Save or Update Parts
        """
        listedParts = []
        retValues = []
        for part in vals:
            hasToBeSaved = False
            if part.get('engineering_code', '') in listedParts:
                continue
            if 'engineering_code' not in part:
                part['componentID'] = False
                part['hasSaved'] = hasToBeSaved
                retValues.append(part)
                continue
            existingID = self.search(cr,uid,[
                                           ('engineering_code', '=', part['engineering_code'])
                                          ,('engineering_revision', '=', part['engineering_revision'])])
            if not existingID:
                new_vals = part.copy()
                try:
                    existingID = self.create(cr, uid, new_vals, context=context)
                except Exception as ex:
                    msg = 'Cannot create product due to error %r\n\n with vals \n%s' % (ex, new_vals)
                    raise Exception(msg)
                hasToBeSaved = True
            else:
                existingID = existingID[0]
                objPart = self.browse(cr, uid, existingID, context=context)
                if not hasToBeSaved:    # Component exists in Odoo
                    part['name'] = objPart.name
                    if self._iswritable(cr, uid, objPart):
                        hasToBeSaved = True
                        del(part['lastupdate'])
                        if not self.write(cr, uid, [existingID], part, context=context):
                            raise osv.except_osv(_('Update Part Error'), _("Part %r cannot be updated" % (part['engineering_code'])))
                        weight = part.get('weight')
                        if weight:
                            if not self.write({'weight': weight}):
                                raise osv.except_osv(_("Part %r cannot be updated" % (part['engineering_code'])))
                        else:
                            logging.warning("No Weight property set unable to update !!")
                
            part['componentID'] = existingID
            part['hasSaved'] = hasToBeSaved
            retValues.append(part)
            listedParts.append(part['engineering_code'])
        return retValues 

    def QueryLast(self, cr, uid, request=([],[]), default=None, context=None):
        """
            Query to return values based on columns selected.
        """
        objId=False
        expData=[]
        queryFilter, columns = request        
        if len(columns)<1:
            return expData
        if 'engineering_revision' in queryFilter:
            del queryFilter['engineering_revision']
        allIDs=self.search(cr,uid,queryFilter,order='engineering_revision',context=context)
        if len(allIDs)>0:
            allIDs.sort()
            objId=allIDs[len(allIDs)-1]
        if objId:
            tmpData=self.export_data(cr, uid, [objId], columns)
            if 'datas' in tmpData:
                expData=tmpData['datas']
        return expData

##  Menu action Methods
    def _create_normalBom(self, cr, uid, idd, context=None):
        """
            Create a new Normal Bom (recursive on all EBom children)
        """
        defaults = {}
        if idd in self.processedIds:
            return False
        checkObj = self.browse(cr, uid, idd, context)
        if not checkObj:
            return False
        bomType = self.pool.get('mrp.bom')
        bomLType = self.pool.get('mrp.bom.line')
        objBoms = bomType.search(cr, uid, [('product_tmpl_id','=',checkObj.product_tmpl_id.id),('type','=','normal')])
        idBoms = bomType.search(cr, uid, [('product_tmpl_id','=',checkObj.product_tmpl_id.id),('type','=','ebom')])

        if not objBoms:
            if idBoms:
                self.processedIds.append(idd)
                newidBom = bomType.copy(cr, uid, idBoms[0], defaults, context)
                if newidBom:
                    bomType.write(cr,uid,[newidBom],{'name':checkObj.engineering_code,'product_id':checkObj.id,'type':'normal',},context=None)
                    oidBom = bomType.browse(cr,uid,newidBom,context=context)
                    ok_rows = self._summarizeBom(cr, uid, oidBom.bom_line_ids)
                    for bom_line in list(set(oidBom.bom_line_ids) ^ set(ok_rows)):
                        bomLType.unlink(cr,uid,[bom_line.id],context=None)
                    for bom_line in ok_rows:
                        bomLType.write(cr,uid,[bom_line.id],{'type':'normal','source_id':False,'name':bom_line.product_id.engineering_code,'product_qty':bom_line.product_qty,},context=None)
                        self._create_normalBom(cr, uid, bom_line.product_id.id, context)
        else:
            for bom_line in bomType.browse(cr, uid, objBoms[0], context=context).bom_line_ids:
                self._create_normalBom(cr, uid, bom_line.product_id.id, context)
        return False

    def _summarizeBom(self, cr, uid, datarows):
        dic = {}
        for datarow in datarows:
            key = '%s_%s' % (datarow.product_id.engineering_code, datarow.product_id.engineering_revision)
            if key in dic:
                dic[key].product_qty = float(dic[key].product_qty) + float(datarow.product_qty)
            else:
                dic[key] = datarow
        retd = dic.values()
        return retd

##  Work Flow Internal Methods
    def _get_recursive_parts(self, cr, uid, ids, excludeStatuses, includeStatuses):
        """
           Get all ids related to current one as children
        """
        stopFlag=False
        tobeReleasedIDs=[]
        if not(type(ids) is types.ListType):
            ids=[ids]
        tobeReleasedIDs.extend(ids)
        children=[]
        for oic in self.browse(cr, uid, ids, context=None):
            children=self.browse(cr, uid, self._getChildrenBom(cr, uid, oic, 1), context=None)
            for child in children:
                if (not child.state in excludeStatuses) and (not child.state in includeStatuses):
                    stopFlag=True
                    break
                if child.state in includeStatuses:
                    if not child.id in tobeReleasedIDs:
                        tobeReleasedIDs.append(child.id)
        return (stopFlag,list(set(tobeReleasedIDs)))

    def action_create_normalBom_WF(self, cr, uid, ids, context=None):
        """
            Create a new Normal Bom if doesn't exist (action callable from code)
        """
        for idd in ids:
            self.processedIds = []
            self._create_normalBom(cr, uid, idd, context)
        self.wf_message_post(cr, uid, ids, body=_('Created Normal Bom.'))
        return False

    def _action_ondocuments(self,cr,uid,ids,action_name,context=None):
        """
            move workflow on documents having the same state of component 
        """
        docIDs=[]
        documentType=self.pool.get('plm.document')
        for oldObject in self.browse(cr, uid, ids, context=context):
            if (action_name!='transmit') and (action_name!='reject') and (action_name!='release'):
                check_state=oldObject.state
            else:
                check_state='confirmed'
            for document in oldObject.linkeddocuments:
                if document.state == check_state:
                    if not document.id in docIDs:
                        if documentType.ischecked_in(cr,uid,document.id,context):
                            docIDs.append(document.id)
        if len(docIDs) > 0:
            if action_name=='confirm':
                documentType.action_confirm(cr,uid,docIDs)
            elif action_name=='transmit':
                documentType.action_confirm(cr,uid,docIDs)
            elif action_name=='draft':
                documentType.action_draft(cr, uid, docIDs)
            elif action_name=='correct':
                documentType.action_draft(cr, uid, docIDs)
            elif action_name=='reject':
                documentType.action_draft(cr, uid, docIDs)
            elif action_name=='release':
                documentType.action_release(cr, uid, docIDs)
            elif action_name=='undermodify':
                documentType.action_cancel(cr,uid,docIDs)
            elif action_name=='suspend':
                documentType.action_suspend(cr,uid,docIDs)
            elif action_name=='reactivate':
                documentType.action_release(cr, uid, docIDs)
            elif action_name=='obsolete':
                documentType.action_obsolete(cr, uid, docIDs)
        return docIDs

    def _iswritable(self, cr, user, oid):
        checkState=('draft')
        if not oid.engineering_writable:
            logging.warning("_iswritable : Part (%r - %d) is not writable." %(oid.engineering_code,oid.engineering_revision))
            return False
        if not oid.state in checkState:
            logging.warning("_iswritable : Part (%r - %d) is in status %r." %(oid.engineering_code,oid.engineering_revision,oid.state))
            return False
        if oid.engineering_code == False:
            logging.warning("_iswritable : Part (%r - %d) is without Engineering P/N." %(oid.name,oid.engineering_revision))
            return False
        return True  

    def wf_message_post_client(self, cr, uid, args, body='', context=None):
        '''
            args = [objId, objMessage]
        '''
        objId, objMessage = args
        self.wf_message_post(cr, uid, [objId], objMessage, context)
        return True

    def wf_message_post(self,cr,uid,ids,body='',context=None):
        """
            Writing messages to follower, on multiple objects
        """
        if not (body==''):
            for id in ids:
                self.message_post(cr, uid, [id], body=_(body))
        
    def action_draft(self,cr,uid,ids,context=None):
        """
            release the object
        """
        defaults={}
        status='draft'
        action='draft'
        docaction='draft'
        defaults['engineering_writable']=True
        defaults['state']=status
        excludeStatuses=['draft','released','undermodify','obsoleted']
        includeStatuses=['confirmed','transmitted']
        return self._action_to_perform(cr, uid, ids, status, action, docaction, defaults, excludeStatuses, includeStatuses, context)

    def action_confirm(self,cr,uid,ids,context=None):
        """
            action to be executed for Draft state
        """
        defaults={}
        status='confirmed'
        action='confirm'
        docaction='confirm'
        defaults['engineering_writable']=False
        defaults['state']=status
        excludeStatuses=['confirmed','transmitted','released','undermodify','obsoleted']
        includeStatuses=['draft']
        return self._action_to_perform(cr, uid, ids, status, action, docaction, defaults, excludeStatuses, includeStatuses, context)

    def action_release(self,cr,uid,ids,context=None):
        """
           action to be executed for Released state
        """
        tmpl_ids=[]
        full_ids=[]
        defaults={}
        prodTmplType=self.pool.get('product.template')
        excludeStatuses=['released','undermodify','obsoleted']
        includeStatuses=['confirmed']
        stopFlag,allIDs=self._get_recursive_parts(cr, uid, ids, excludeStatuses, includeStatuses)
        if len(allIDs)<1 or stopFlag:
            raise osv.except_osv(_('WorkFlow Error'), _("Part cannot be released."))
        allProdObjs=self.browse(cr, uid, allIDs, context=context)
        for oldObject in allProdObjs:
            last_id=self._getbyrevision(cr, uid, oldObject.engineering_code, oldObject.engineering_revision-1)
            if last_id != None:
                defaults['engineering_writable']=False
                defaults['state']='obsoleted'
                prodObj=self.browse(cr, uid, [last_id], context=context)
                prodTmplType.write(cr,uid,[prodObj.product_tmpl_id.id],defaults ,context=context)
                self.wf_message_post(cr, uid, [last_id], body=_('Status moved to: %s.' %(USEDIC_STATES[defaults['state']])))
            defaults['engineering_writable']=False
            defaults['state']='released'
        self._action_ondocuments(cr,uid,allIDs,'release')
        for currId in allProdObjs:
            if not(currId.id in ids):
                tmpl_ids.append(currId.id)
            full_ids.append(currId.product_tmpl_id.id)
        self.signal_workflow(cr, uid, tmpl_ids, 'release')
        objId=self.pool.get('product.template').write(cr, uid, full_ids, defaults, context=context)
        if (objId):
            self.wf_message_post(cr, uid, allIDs, body=_('Status moved to: %s.' %(USEDIC_STATES[defaults['state']])))
        return objId

    def action_obsolete(self,cr,uid,ids,context=None):
        """
            obsolete the object
        """
        defaults={}
        status='obsoleted'
        action='obsolete'
        docaction='obsolete'
        defaults['engineering_writable']=False
        defaults['state']=status
        excludeStatuses=['draft','confirmed','transmitted','undermodify','obsoleted']
        includeStatuses=['released']
        return self._action_to_perform(cr, uid, ids, status, action, docaction, defaults, excludeStatuses, includeStatuses, context)

    def action_reactivate(self,cr,uid,ids,context=None):
        """
            reactivate the object
        """
        defaults={}
        status='released'
        action=''
        docaction='release'
        defaults['engineering_writable']=True
        defaults['state']=status
        excludeStatuses=['draft','confirmed','transmitted','released','undermodify','obsoleted']
        includeStatuses=['obsoleted']
        return self._action_to_perform(cr, uid, ids, status, action, docaction, defaults, excludeStatuses, includeStatuses, context)
    
#   WorkFlow common internal method to apply changes

    def _action_to_perform(self, cr, uid, ids, status, action, docaction, defaults=[], excludeStatuses=[], includeStatuses=[], context=None):
        tmpl_ids=[]
        full_ids=[]
        stopFlag,allIDs=self._get_recursive_parts(cr, uid, ids, excludeStatuses, includeStatuses)
        self._action_ondocuments(cr,uid,allIDs,docaction)
        for currId in self.browse(cr,uid,allIDs,context=context):
            if not(currId.id in ids):
                tmpl_ids.append(currId.id)
            full_ids.append(currId.product_tmpl_id.id)
        if action:
            self.signal_workflow(cr, uid, tmpl_ids, action)
        objId=self.pool.get('product.template').write(cr, uid, full_ids, defaults, context=context)
        if (objId):
            self.wf_message_post(cr, uid, allIDs, body=_('Status moved to: %s.' %(USEDIC_STATES[defaults['state']])))
        return objId
    
#######################################################################################################################################33

#   Overridden methods for this entity

    def create(self, cr, uid, vals, context=None):
        if not vals:
            return False
        if ('name' in vals):
            if not vals['name']:
                return False
            existingIDs=self.search(cr, uid, [('name','=',vals['name'])], order = 'engineering_revision', context=context)
            if 'engineering_code' in vals:
                if vals['engineering_code'] == False:
                    vals['engineering_code'] = vals['name']
            else:
                vals['engineering_code'] = vals['name']
    
            if existingIDs:
                existingID=existingIDs[len(existingIDs)-1]
                if ('engineering_revision' in vals):
                    existObj=self.browse(cr,uid,existingID,context=context)
                    if existObj:
                        if vals['engineering_revision'] > existObj.engineering_revision:
                            vals['name']=existObj.name
                        else:
                            return existingID
                else:
                    return existingID

        vals.update(self.checkMany2oneClient(cr, uid, [], vals))
        logging.info('[plm] Going to create product with values %r' % (vals))
        return super(plm_component, self).create(cr, uid, vals, context=context)

    def copy(self, cr, uid, id, default={}, context=None):
        """
            Overwrite the default copy method
        """
        logging.info('Going to clone self %r, id to clone %r, defaults %r' % (self, id, default))
        previous_name = self.browse(cr, uid, id, context=context).name
        if not default.get('name', False):
            default['name'] = '-'
            default['engineering_code'] = '-'
            default['engineering_revision'] = 0
        # assign default value
        default['state'] = 'draft'
        default['engineering_writable'] = True
        default['write_date'] = None
        default['linkeddocuments'] = []
        logging.info('Copy overload with values %r' % (default))
        objId = super(plm_component, self).copy(cr, uid, id, default, context=context)
        if (objId):
            self.wf_message_post(cr, uid, [id], body=_('Copied starting from : %s.' % previous_name))
        return objId

    @api.multi
    def read(self, fields=[], load='_classic_read'):
        if not fields:
            fields = []
        customFields = [field.replace('plm_m2o_', '') for field in fields if field.startswith('plm_m2o_')]
        fields.extend(customFields)
        fields = list(set(fields))
        logging.info('Read plm component %r with values %r and load %r' % (self.ids, fields, load))
        res = super(plm_component, self).read(fields=fields, load=load)
        res = self.readMany2oneFields(res, fields)
        return res

    @api.multi
    def readMany2oneFields(self, readVals, fields):
        out = []
        for vals in readVals:
            fields_def = self.fields_get(vals.keys())
            tmpVals = vals.copy()
            for fieldName, fieldVal in vals.items():
                customField = 'plm_m2o_' + fieldName
                if customField in fields:
                    logging.info('Reading many2one field %r, fieldVal %r' % (customField, fieldVal))
                    field_def = fields_def.get(fieldName)
                    if field_def.get('type', '') == 'many2one':
                        relation = field_def.get('relation', '')
                        tmpVals[customField] = ''
                        if relation:
                            if isinstance(fieldVal, int):
                                tmpVals[customField] = self.env[relation].browse(fieldVal).name
                            elif isinstance(fieldVal, (list, tuple)):
                                tmpVals[customField] = fieldVal[1]
                            else:
                                tmpVals[customField] = ''
            out.append(tmpVals)
        return out

    @api.multi
    def write(self, vals):
        if not 'is_engcode_editable' in vals:
            vals['is_engcode_editable'] = False
        vals.update(self.checkMany2oneClient(vals))
        return super(plm_component, self).write(vals)

    @api.multi
    def checkMany2oneClient(self, vals):
        out = {}
        customFields = [field.replace('plm_m2o_', '') for field in vals.keys() if field.startswith('plm_m2o_')]
        fieldsGet = self.fields_get(customFields)
        for fieldName, fieldDefinition in fieldsGet.items():
            refId = self.customFieldConvert(fieldDefinition, vals, fieldName)
            if refId:
                out[fieldName] = refId.id
        return out

    @api.multi
    def customFieldConvert(self, fieldDefinition, vals, fieldName):
        refId = False
        fieldType = fieldDefinition.get('type', '')
        referredModel = fieldDefinition.get('relation', '')
        oldFieldName = 'plm_m2o_' + fieldName
        cadVal = vals.get(oldFieldName, '')
        if fieldType == 'many2one':
            try:
                for refId in self.env[referredModel].search([('name', '=', cadVal)]):
                    return refId
            except Exception as ex:
                logging.warning(ex)
        return refId

    def unlink(self, cr, uid, ids, context=None):
        values={'state':'released',}
        checkState=('undermodify','obsoleted')
        for checkObj in self.browse(cr, uid, ids, context=context):
            existingID = self.search(cr, uid, [('engineering_code', '=', checkObj.engineering_code),('engineering_revision', '=', checkObj.engineering_revision-1)])
            if len(existingID)>0:
                oldObject=self.browse(cr, uid, existingID[0], context=context)
                if oldObject.state in checkState:
                    self.wf_message_post(cr, uid, [oldObject.id], body=_('Removed : Latest Revision.'))
                    context['check'] = False
                    if not self.write(cr, uid, [oldObject.id], values, context):
                        logging.warning("unlink : Unable to update state to old component (%r - %d)." %(oldObject.engineering_code,oldObject.engineering_revision))
                        return False
        return super(plm_component,self).unlink(cr, uid, ids, context=context)

#   Overridden methods for this entity

    def translateForClient(self, cr, uid, values=[], forcedLang='', context={}):
        '''
            Get values attribute in this format:
            values = [{'field1':value1,'field2':value2,...}]     only one element in the list!!!
            and return computed values due to language
            
            Get also forcedLang attribute in this format:
            forcedLang = 'en_US'
            if is not set it takes language from user
        '''
        language = forcedLang
        if not forcedLang:
            resDict = self.pool.get('res.users').read(cr, uid, uid, ['lang'])
            language = resDict.get('lang','')
        if values:
            values = values[0]
        if language and values:
            toRead = filter(lambda x: type(x) in [str, unicode] and x,values.values()) # Where computed only string and not null string values (for performance improvement)
            toRead = list(set(toRead))                                                 # Remove duplicates
            for fieldName, valueToTranslate in values.items():
                if valueToTranslate not in toRead:
                    continue
                translationObj = self.pool.get('ir.translation')
                resIds = translationObj.search(cr, uid, [('lang','=',language),('src','=',valueToTranslate)])
                if resIds:
                    readDict = translationObj.read(cr, uid, resIds[0], ['value'])
                    values[fieldName] = readDict.get('value','')
        return values
    
    def action_rev_docs(self, cr, uid, ids, context={}):
        '''
            This function is called by the button on component view, section LinkedDocuments
            Clicking that button all documents related to all revisions of this component are opened in a tree view
        '''
        docIds = []
        for compBrws in self.browse(cr, uid, ids, context):
            engineering_code = compBrws.engineering_code
            if not engineering_code:
                logging.warning("Part %s doesn't have and engineering code!" %(compBrws.name))
                continue
            componentIds = self.search(cr, uid, [('engineering_code', '=' , engineering_code)])
            for compBrws in self.browse(cr, uid, componentIds, context):
                docIds.extend(compBrws.linkeddocuments.ids)
        return {
            'domain': [('id', 'in', docIds)],
            'name': _('Related documents'),
            'view_type': 'form',
            'view_mode': 'tree,form',
            'res_model': 'plm.document',
            'type': 'ir.actions.act_window',
         }

    @api.multi
    def name_get(self):
        result = []
        for inv in self:
            newName = "%s [Rev %r] %s" % (inv.engineering_code, inv.engineering_revision, inv.name)
            result.append((inv.id, newName))
        return result

    def isLastDocumentRevision(self, cr, uid, compId, docId, context={}):
        '''
            NOT USED
            Verify if document is at it's last revision
            I can't do a simple search ordering by revisionid because document names changes every time by sequence
        '''
        if not docId:
            msg = 'Wrong values on verification if is selected last document revision: compId="%r", docId="%r"' % (compId, docId)
            logging.error(msg)
            raise msg
        compBrws = self.browse(cr, uid, compId)
        compIds = self.search(cr, uid, [('engineering_code', '=', compBrws.engineering_code)], order='engineering_revision')
        lastcompBrws = self.browse(cr, uid, compIds[-1])
        docBrwstoOpen = self.pool.get('plm.document').browse(cr, uid, docId)
        for docBrws in lastcompBrws.linkeddocuments:
            if docBrws.revisionid > docBrwstoOpen.revisionid:
                # This verification is because I may have multiple document revisions in the same last revision component
                return False
        return True

    @api.model
    def ConvertToPlmProduct(self, attributes_list=[]):
        """
        convert the attributes vals and add the plm id to the array
        """
        out = []
        for attributes in attributes_list:
            engineering_code = attributes.get('engineering_code', False)
            engineering_revision = attributes.get('engineering_revision', 0)
            if engineering_code:
                product_product_id = self.search([('engineering_code', '=', engineering_code),
                                                  ('engineering_revision', '=', engineering_revision)], order='engineering_revision ASC')
                if not product_product_id:
                    product_product_id = self.create(attributes)
                attributes['plm_id'] = product_product_id.id
                out.append(attributes)
        return out

plm_component()
